package at.bestsolution.rsd.rt.json;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.OptionalInt;
import java.util.OptionalLong;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Function;

import javax.json.stream.JsonGenerator;
import javax.json.stream.JsonParser;
import javax.json.stream.JsonParser.Event;

import at.bestsolution.rsd.rt.Array;
import at.bestsolution.rsd.rt.Delta;
import at.bestsolution.rsd.rt.RSDBuilder;
import at.bestsolution.rsd.rt.Ref;
import at.bestsolution.rsd.rt.delta.BooleanValue;
import at.bestsolution.rsd.rt.delta.DoubleValue;
import at.bestsolution.rsd.rt.delta.IntValue;
import at.bestsolution.rsd.rt.delta.ListChange;
import at.bestsolution.rsd.rt.delta.LongValue;
import at.bestsolution.rsd.rt.delta.Value;
import at.bestsolution.rsd.rt.delta.ValueChange;

public class JsonStreamUtil {
    static final String ID = "id";
    static final String ID_REF = "idRef";
    static final String ETAG = "etag";
    static final String VALUE = "value";

    public static class Read {
        public static void skipSubtree(JsonParser parser) {
            Event next = parser.next();
            if( next == Event.START_OBJECT ) {
                parser.skipObject();
            } else if( next == Event.START_ARRAY ) {
                parser.skipArray();
            }
        }
        
        public static Optional<String> readString(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                JsonParser.Event e = parser.next();
                if (e == JsonParser.Event.VALUE_NULL) {
                    return Optional.empty();
                } else if (e == JsonParser.Event.VALUE_STRING) {
                    return Optional.of(parser.getString());
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a String", keyName));
        }

        public static Value<String> readValueString(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                JsonParser.Event e = parser.next();
                if (e == JsonParser.Event.VALUE_NULL) {
                    return Value.NULL();
                } else if (e == JsonParser.Event.VALUE_STRING) {
                    return Value.of(parser.getString());
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a String", keyName));
        }

        public static Optional<LocalDate> readLocalDate(String keyName, JsonParser parser) {
            return readString(keyName, parser).map(LocalDate::parse);
        }

        public static Value<LocalDate> readValueLocalDate(String keyName, JsonParser parser) {
            return readValueString(keyName, parser).map(LocalDate::parse);
        }

        public static Optional<LocalTime> readLocalTime(String keyName, JsonParser parser) {
            return readString(keyName, parser).map(LocalTime::parse);
        }

        public static Value<LocalTime> readValueLocalTime(String keyName, JsonParser parser) {
            return readValueString(keyName, parser).map(LocalTime::parse);
        }

        public static Optional<LocalDateTime> readLocalDateTime(String keyName, JsonParser parser) {
            return readString(keyName, parser).map(LocalDateTime::parse);
        }

        public static Value<LocalDateTime> readValueLocalDateTime(String keyName, JsonParser parser) {
            return readValueString(keyName, parser).map(LocalDateTime::parse);
        }

        public static Optional<ZonedDateTime> readZonedDateTime(String keyName, JsonParser parser) {
            return readString(keyName, parser).map(ZonedDateTime::parse);
        }

        public static Optional<Boolean> readBoolean(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_TRUE) {
                    return Optional.of(Boolean.TRUE);
                } else if (event == Event.VALUE_FALSE) {
                    return Optional.of(Boolean.FALSE);
                } else if (event == Event.VALUE_NULL) {
                    return Optional.empty();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a boolean", keyName));
        }

        public static BooleanValue readValueBoolean(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_TRUE) {
                    return BooleanValue.TRUE();
                } else if (event == Event.VALUE_FALSE) {
                    return BooleanValue.FALSE();
                } else if (event == Event.VALUE_NULL) {
                    return BooleanValue.NULL();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a boolean", keyName));
        }

        public static OptionalInt readInt(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return OptionalInt.of(parser.getInt());
                } else if (event == Event.VALUE_NULL) {
                    return OptionalInt.empty();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a int", keyName));
        }

        public static IntValue readValueInt(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return IntValue.of(parser.getInt());
                } else if (event == Event.VALUE_NULL) {
                    return IntValue.NULL();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a int", keyName));
        }

        public static OptionalLong readLong(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return OptionalLong.of(parser.getLong());
                } else if (event == Event.VALUE_NULL) {
                    return OptionalLong.empty();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a long", keyName));
        }

        public static LongValue readValueLong(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return LongValue.of(parser.getLong());
                } else if (event == Event.VALUE_NULL) {
                    return LongValue.NULL();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a long", keyName));
        }

        public static OptionalDouble readDouble(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return OptionalDouble.of(parser.getBigDecimal().doubleValue());
                } else if (event == Event.VALUE_NULL) {
                    return OptionalDouble.empty();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a double", keyName));
        }

        public static DoubleValue readValueDouble(String keyName, JsonParser parser) {
            if (parser.hasNext()) {
                Event event = parser.next();
                if (event == Event.VALUE_NUMBER) {
                    return DoubleValue.of(parser.getBigDecimal().doubleValue());
                } else if (event == Event.VALUE_NULL) {
                    return DoubleValue.NULL();
                }
            }
            throw new IllegalStateException(String.format("Property '%s' is not a double", keyName));
        }

        public static <T> Ref<T> readRefArrayElement(String keyName, Class<T> clazz, JsonParser parser) {
            String idRef = null;
            String etag = null;

            while (parser.hasNext()) {
                Event event = parser.next();
                switch (event) {
                case START_OBJECT:
                    break;
                case KEY_NAME:
                    if (ID_REF.equals(parser.getString())) {
                        idRef = readString(ID_REF, parser)
                                .orElseThrow(() -> new IllegalStateException("idRef must not be null"));
                    } else if (ETAG.equals(parser.getString())) {
                        etag = readString(ETAG, parser).orElse(null);
                    }
                    break;
                case END_OBJECT:
                    if (idRef == null) {
                        throw new IllegalStateException("Ref needs an idRef-Attribute");
                    }
                    return Ref.of(clazz, idRef, etag);
                default:
                    break;
                }
            }

            throw new IllegalStateException(String.format("Invalid ref definition for property '%s'", keyName));
        }

        public static <T> Optional<Ref<T>> readRef(String keyName, Class<T> clazz, JsonParser parser) {
            if (!parser.hasNext()) {
                throw new IllegalStateException(String.format("Invalid state for property '%s'", keyName));
            } else {
                Event event = parser.next();
                if (event == Event.VALUE_NULL) {
                    return Optional.empty();
                }
            }

            String idRef = null;
            String etag = null;

            while (parser.hasNext()) {
                Event event = parser.next();
                switch (event) {
                case START_OBJECT:
                    break;
                case KEY_NAME:
                    if (ID_REF.equals(parser.getString())) {
                        idRef = readString(ID_REF, parser)
                                .orElseThrow(() -> new IllegalStateException("idRef must not be null"));
                    } else if (ETAG.equals(parser.getString())) {
                        etag = readString(ETAG, parser).orElse(null);
                    }
                    break;
                case END_OBJECT:
                    if (idRef == null) {
                        throw new IllegalStateException("Ref needs an idRef-Attribute");
                    }
                    return Optional.of(Ref.of(clazz, idRef, etag));
                default:
                    break;
                }
            }

            throw new IllegalStateException(String.format("Invalid ref definition for property '%s'", keyName));
        }

        public static <T> Optional<Optional<T>> readValueChangeObject(String keyName, JsonParser parser,
                BiFunction<String, JsonParser, Optional<T>> valueHandler) {
            Optional<T> value = null;

            while (parser.hasNext()) {
                Event event = parser.next();

                switch (event) {
                case START_OBJECT:
                    break;
                case VALUE_NULL:
                    return Optional.empty();
                case KEY_NAME:
                    if (VALUE.equals(parser.getString())) {
                        value = valueHandler.apply(keyName, parser);
                    }
                    break;
                case END_OBJECT:
                    if (value == null) {
                        throw new IllegalStateException(String.format("Invalid ChangeValue object for %s", keyName));
                    }
                    return Optional.of(value);
                default:
                    break;
                }
            }
            throw new IllegalStateException(String.format("Invalid ChangeValue object for %s", keyName));
        }

        public static <T> Optional<List<T>> readArray(String name, JsonParser parser,
                Function<JsonParser, T> valueHandler) {
            List<T> rv = null;

            while (parser.hasNext()) {
                switch (parser.next()) {
                case VALUE_NULL:
                    if (rv != null) {
                        throw new IllegalStateException(
                                String.format("Null objects not allowed in list of '%s'", name));
                    }
                    return Optional.empty();
                case START_ARRAY:
                    if (rv != null) {
                        throw new IllegalStateException(String.format("Invalid sublist in %s", name));
                    }
                    rv = new ArrayList<>();
                    break;
                case END_ARRAY:
                    return Optional.of(rv);
                case START_OBJECT:
                    if (rv == null) {
                        throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                    }
                    rv.add(valueHandler.apply(parser));
                    break;
                default:
                    throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                }
            }

            return Optional.of(rv);
        }

        public static Optional<String[]> readStringArray(String name, JsonParser parser) {
            List<String> rv = null;

            while (parser.hasNext()) {
                switch (parser.next()) {
                case VALUE_NULL:
                    throw new IllegalStateException(String.format("Null objects not allowed in list of '%s'", name));
                case START_ARRAY:
                    if (rv != null) {
                        throw new IllegalStateException(String.format("Invalid sublist in %s", name));
                    }
                    rv = new ArrayList<>();
                    break;
                case END_ARRAY:
                    return Optional.of(rv.toArray(new String[0]));
                case START_OBJECT:
                    throw new IllegalStateException(String.format("Objects not allowed in list of '%s'", name));
                case VALUE_STRING:
                    if (rv == null) {
                        throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                    }
                    rv.add(parser.getString());
                    break;
                default:
                    throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                }
            }

            return Optional.of(rv.toArray(new String[0]));
        }

        public static Optional<double[]> readDoubleArray(String name, JsonParser parser) {
            List<Double> rv = null;

            while (parser.hasNext()) {
                switch (parser.next()) {
                case VALUE_NULL:
                    throw new IllegalStateException(String.format("Null objects not allowed in list of '%s'", name));
                case START_ARRAY:
                    if (rv != null) {
                        throw new IllegalStateException(String.format("Invalid sublist in %s", name));
                    }
                    rv = new ArrayList<>();
                    break;
                case END_ARRAY:
                    return Optional.of(rv.stream().mapToDouble(Double::doubleValue).toArray());
                case START_OBJECT:
                    throw new IllegalStateException(String.format("Objects not allowed in list of '%s'", name));
                case VALUE_NUMBER:
                    if (rv == null) {
                        throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                    }
                    rv.add(parser.getBigDecimal().doubleValue());
                    break;
                default:
                    throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                }
            }

            return Optional.of(rv.stream().mapToDouble(Double::doubleValue).toArray());
        }

        public static Optional<int[]> readIntArray(String name, JsonParser parser) {
            List<Integer> rv = null;

            while (parser.hasNext()) {
                switch (parser.next()) {
                case VALUE_NULL:
                    throw new IllegalStateException(String.format("Null objects not allowed in list of '%s'", name));
                case START_ARRAY:
                    if (rv != null) {
                        throw new IllegalStateException(String.format("Invalid sublist in %s", name));
                    }
                    rv = new ArrayList<>();
                    break;
                case END_ARRAY:
                    return Optional.of(rv.stream().mapToInt(Integer::intValue).toArray());
                case START_OBJECT:
                    throw new IllegalStateException(String.format("Objects not allowed in list of '%s'", name));
                case VALUE_NUMBER:
                    if (rv == null) {
                        throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                    }
                    rv.add(parser.getInt());
                    break;
                default:
                    throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                }
            }

            return Optional.of(rv.stream().mapToInt(Integer::intValue).toArray());
        }

        public static Optional<long[]> readLongArray(String name, JsonParser parser) {
            List<Long> rv = null;

            while (parser.hasNext()) {
                switch (parser.next()) {
                case VALUE_NULL:
                    throw new IllegalStateException(String.format("Null objects not allowed in list of '%s'", name));
                case START_ARRAY:
                    if (rv != null) {
                        throw new IllegalStateException(String.format("Invalid sublist in %s", name));
                    }
                    rv = new ArrayList<>();
                    break;
                case END_ARRAY:
                    return Optional.of(rv.stream().mapToLong(Long::longValue).toArray());
                case START_OBJECT:
                    throw new IllegalStateException(String.format("Objects not allowed in list of '%s'", name));
                case VALUE_NUMBER:
                    if (rv == null) {
                        throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                    }
                    rv.add(parser.getLong());
                    break;
                default:
                    throw new IllegalStateException(String.format("Invalid multi-value property %s", name));
                }
            }

            return Optional.of(rv.stream().mapToLong(Long::longValue).toArray());
        }

        public static <T extends Delta> Optional<T> readDelta(JsonParser parser,
                BiConsumer<JsonParser, String> propertyHandler, Delta.DeltaBuilder<T> builder) {
            String id = null;
            String etag = null;
            while (parser.hasNext()) {
                switch (parser.next()) {
                case START_OBJECT:
                    break;
                case VALUE_NULL:
                    return Optional.empty();
                case END_OBJECT:
                    return Optional.of(builder.build(id, etag));
                case KEY_NAME:
                    if (ID.equals(parser.getString())) {
                        id = readString(ID, parser)
                                .orElseThrow(() -> new IllegalStateException("A delta object has to have an ID"));
                    } else if (ETAG.equals(parser.getString())) {
                        etag = readString(ETAG, parser).orElse(null);
                    } else {
                        propertyHandler.accept(parser, parser.getString());
                    }
                    break;
                default:
                    throw new IllegalStateException("Unsupported event");
                }
            }

            throw new IllegalStateException("Unsupported event");
        }

        public static <T> Optional<T> readRSDObject(JsonParser parser, BiConsumer<JsonParser, String> propertyHandler,
                RSDBuilder<T> builder) {
            while (parser.hasNext()) {
                JsonParser.Event event = parser.next();
                switch (event) {
                case START_OBJECT:
                    break;
                case VALUE_NULL:
                    return Optional.empty();
                case END_OBJECT:
                    return Optional.of(builder.build());
                case KEY_NAME:
                    propertyHandler.accept(parser, parser.getString());
                    break;
                default:
                    throw new IllegalStateException(String.format("Unsupported event '%s'", event));
                }
            }
            throw new IllegalStateException("Unsupported event");

        }

        public static <O> ListChange<O, Ref<O>> readObjectListChange(JsonParser parser,
                Function<JsonParser, O> valueHandler, Class<?> clazz) {
            ListChange.Builder<O, Ref<O>> b = new ListChange.Builder<>();
            Event event = parser.next();

            if (event == Event.START_OBJECT) {
                while (parser.hasNext()) {
                    switch (parser.next()) {
                    case KEY_NAME: {
                        String key = parser.getString();
                        if ("additions".equals(key)) {
                            handleAdditions(parser, b, valueHandler);
                        } else if ("removals".equals(key)) {
                            handleRemovals(parser, b, clazz);
                        } else {
                            throw new IllegalStateException(String.format("Unknown list change '%s'", key));
                        }
                        break;
                    }
                    case END_OBJECT:
                        return b.build();
                    default:
                        break;
                    }
                }
            } else {
                throw new IllegalStateException("Change object not started");
            }

            return ListChange.EMPTY();
        }

        private static <O> void handleRemovals(JsonParser parser, ListChange.Builder<O, Ref<O>> b, Class<?> clazz) {
            int idx = -1;
            Array<Ref<O>> elements = Array.empty();
            while (parser.hasNext()) {
                switch (parser.next()) {
                case START_OBJECT:
                    idx = -1;
                    elements = Array.empty();
                    break;
                case KEY_NAME: {
                    String key = parser.getString();
                    if ("index".equals(key)) {
                        if (parser.next() == Event.VALUE_NUMBER) {
                            idx = parser.getInt();
                        } else {
                            throw new IllegalStateException("Add index must be a number");
                        }
                    } else if ("elements".equals(key)) {
                        elements = Array.of(readArray("elements", parser, p -> {
                            return (Ref<O>)readRefArrayElement(key, clazz, p);    
                        }).orElse(Collections.emptyList()));
                    } else {
                        throw new IllegalStateException(String.format("Unknown Remove-Attribute %s", key));
                    }
                }
                case END_OBJECT:
                    if (idx == -1) {
                        b.remove(elements);
                    } else {
                        b.remove(idx, elements);
                    }
                    break;
                case END_ARRAY:
                    return;
                default:
                    break;
                }
            }
        }
        
        private static <O> void handleAdditions(JsonParser parser, ListChange.Builder<O, Ref<O>> b,
                Function<JsonParser, O> valueHandler) {
            int idx = -1;
            Array<O> elements = Array.empty();
            while (parser.hasNext()) {
                switch (parser.next()) {
                case START_OBJECT:
                    idx = -1;
                    elements = Array.empty();
                    break;
                case KEY_NAME: {
                    String key = parser.getString();
                    if ("index".equals(key)) {
                        if (parser.next() == Event.VALUE_NUMBER) {
                            idx = parser.getInt();
                        } else {
                            throw new IllegalStateException("Add index must be a number");
                        }
                    } else if ("elements".equals(key)) {
                        elements = Array
                                .of(readArray("elements", parser, valueHandler).orElse(Collections.emptyList()));
                    } else {
                        throw new IllegalStateException(String.format("Unknown Add-Attribute %s", key));
                    }
                    break;
                }
                case END_OBJECT:
                    if (idx == -1) {
                        b.append(elements);
                    } else {
                        b.insert(idx, elements);
                    }
                    break;
                case END_ARRAY:
                    return;
                default:
                    break;
                }
            }
        }

//		@Deprecated
//		public static <T extends RSDOject> Optional<T> readRSDObject(JsonParser parser, BiConsumer<JsonParser, String> propertyHandler, RSDOject.RSDOjectBuilder<T> builder) {
//			while( parser.hasNext() ) {
//				switch (parser.next()) {
//				case START_OBJECT:
//					break;
//				case VALUE_NULL:
//					return Optional.empty();
//				case END_OBJECT:
//					return Optional.of(builder.build());
//				case KEY_NAME:
//					if( ID.equals(parser.getString()) ) {
//						builder.id(readString(ID, parser).orElseThrow( () -> new IllegalStateException("A delta object has to have an ID")));
//					} else if( ETAG.equals(parser.getString()) ) {
//						builder.etag(readString(ETAG, parser).orElse(null));
//					} else {
//						propertyHandler.accept(parser, parser.getString());
//					}
//					break;
//				default:
//					throw new IllegalStateException("Unsupported event");
//				}
//			}
//			throw new IllegalStateException("Unsupported event");
//		}
    }

    public static class Write {
        public static void writeString(String keyName, String value, JsonGenerator generator) {
            if (value != null) {
                generator.write(keyName, value);
            }
        }

        public static void writeValueString(String keyName, Value<String> value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.value());
                }
            }
        }

        public static void writeString(String value, JsonGenerator generator) {
            if (value != null) {
                generator.write(value);
            } else {
                generator.writeNull();
            }
        }

        public static void writeLocalDate(String keyName, LocalDate value, JsonGenerator generator) {
            if (value != null) {
                generator.write(keyName, value.toString());
            }
        }

        public static void writeValueLocalDate(String keyName, Value<LocalDate> value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.value().toString());
                }
            }
        }

        public static void writeLocalDate(LocalDate value, JsonGenerator generator) {
            if (value != null) {
                generator.write(value.toString());
            } else {
                generator.writeNull();
            }
        }

        public static void writeLocalTime(String keyName, LocalTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(keyName, value.toString());
            }
        }

        public static void writeValueLocalTime(String keyName, Value<LocalTime> value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.value().toString());
                }
            }
        }

        public static void writeLocalTime(LocalTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(value.toString());
            } else {
                generator.writeNull();
            }
        }

        public static void writeLocalDateTime(String keyName, LocalDateTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(keyName, value.toString());
            }
        }

        public static void writeValueLocalDateTime(String keyName, Value<LocalDateTime> value,
                JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.value().toString());
                }
            }
        }

        public static void writeLocalDateTime(LocalDateTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(value.toString());
            } else {
                generator.writeNull();
            }
        }

        public static void writeZonedDateTime(String keyName, ZonedDateTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(keyName, value.toString());
            }
        }

        public static void writeValueZonedDateTime(String keyName, Value<ZonedDateTime> value,
                JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.value().toString());
                }
            }
        }

        public static void writeZonedDateTime(ZonedDateTime value, JsonGenerator generator) {
            if (value != null) {
                generator.write(value.toString());
            } else {
                generator.writeNull();
            }
        }

        public static void writeBoolean(String keyName, boolean value, JsonGenerator generator) {
            generator.writeKey(keyName);
            generator.write(value);
        }

        public static void writeValueBoolean(String keyName, BooleanValue value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.booleanValue());
                }
            }
        }

        public static void writeBoolean(String keyName, Boolean value, JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                generator.write(value);
            }
        }

        public static void writeBoolean(boolean value, JsonGenerator generator) {
            generator.write(value);
        }

        public static void writeInt(String keyName, int value, JsonGenerator generator) {
            generator.writeKey(keyName);
            generator.write(value);
        }

        public static void writeValueInt(String keyName, IntValue value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.intValue());
                }
            }
        }

        public static void writeInt(String keyName, Integer value, JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                generator.write(value);
            }
        }

        public static void writeInt(int value, JsonGenerator generator) {
            generator.write(value);
        }

        public static void writeLong(String keyName, long value, JsonGenerator generator) {
            generator.writeKey(keyName);
            generator.write(value);
        }

        public static void writeValueLong(String keyName, LongValue value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.longValue());
                }
            }
        }

        public static void writeLong(String keyName, Long value, JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                generator.write(value);
            }
        }

        public static void writeLong(long value, JsonGenerator generator) {
            generator.write(value);
        }

        public static void writeDouble(String keyName, double value, JsonGenerator generator) {
            generator.writeKey(keyName);
            generator.write(value);
        }

        public static void writeValueDouble(String keyName, DoubleValue value, JsonGenerator generator) {
            if (value.isDefined()) {
                if (value.isNull()) {
                    generator.writeNull(keyName);
                } else {
                    generator.write(keyName, value.doubleValue());
                }
            }
        }

        public static void writeDouble(String keyName, Double value, JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                generator.write(value);
            }
        }

        public static void writeDouble(double value, JsonGenerator generator) {
            generator.write(value);
        }

        public static void writeRef(Ref<?> value, JsonGenerator generator) {
            if (value != null) {
                generator.writeStartObject();
                generator.write(ID_REF, value.id());
                if (value.etag() != null) {
                    generator.write(ETAG, value.etag());
                }
                generator.writeEnd();
            } else {
                generator.writeNull();
            }
        }

        public static void writeRef(String keyName, Ref<?> value, JsonGenerator generator) {
            if (value != null) {
                generator.writeStartObject(keyName);
                generator.write(ID_REF, value.id());
                if (value.etag() != null) {
                    generator.write(ETAG, value.etag());
                }
                generator.writeEnd();
            }
        }

        public static <T> void writeValueChangeObject(String keyName, Optional<ValueChange<T>> value,
                BiConsumer<T, JsonGenerator> valueHandler, JsonGenerator generator) {
            if (value.isPresent()) {
                generator.writeStartObject(keyName);
                if (value.get().value() != null) {
                    generator.writeKey(VALUE);
                    valueHandler.accept(value.get().value(), generator);
                } else {
                    generator.writeNull(VALUE);

                }
                generator.writeEnd();
            }
        }

        public static <T extends Delta> void writeDeltaObject(T value, BiConsumer<T, JsonGenerator> propertiesHandler,
                JsonGenerator generator) {
            if (value != null) {
                generator.writeStartObject();
                if (value.id() != null) {
                    generator.write(ID, value.id());
                }
                if (value.etag() != null) {
                    generator.write(ETAG, value.etag());
                }
                propertiesHandler.accept(value, generator);
                generator.writeEnd();
            }
        }

        public static <T> void writeObject(String keyName, T value, BiConsumer<T, JsonGenerator> propertiesHandler,
                JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                writeObject(value, propertiesHandler, generator);
            }
        }

        public static <T> void writeObject(T value, BiConsumer<T, JsonGenerator> propertiesHandler,
                JsonGenerator generator) {
            if (value != null) {
                generator.writeStartObject();
                propertiesHandler.accept(value, generator);
                generator.writeEnd();
            } else {
                generator.writeNull();
            }
        }

        public static <O> void writeObjectListChange(ListChange<O, Ref<O>> change,
                BiConsumer<Array<O>, JsonGenerator> objSerializer, JsonGenerator writer) {
            writer.writeStartObject();

            if (!change.additions().isEmpty()) {
                writer.writeKey("additions");
                writer.writeStartArray();
                change.additions().forEach(add -> {
                    writer.writeStartObject();
                    add.index().ifPresent(idx -> writer.write("index", idx));
                    writer.writeKey("elements");
                    objSerializer.accept(add.elements(), writer);
                    writer.writeEnd();
                });
                writer.writeEnd();
            }

            if (!change.removals().isEmpty()) {
                writer.writeKey("removals");
                writer.writeStartArray();
                change.removals().forEach(remove -> {
                    writer.writeStartObject();
                    remove.index().ifPresent(idx -> writer.write("index", idx));
                    writer.writeKey("elements");
                    writer.writeStartArray();
                    remove.elements().forEach(ref -> writeRef(ref, writer));
                    writer.writeEnd();
                    writer.writeEnd();
                });
                writer.writeEnd();
            }

            writer.writeEnd();
        }

//		@Deprecated
//		public static <T extends RSDOject> void writeRSDObject(String keyName, T value, BiConsumer<T,JsonGenerator> propertiesHandler, JsonGenerator generator) {
//			if( value != null ) {
//				generator.writeKey(keyName);
//				writeRSDObject(value, propertiesHandler, generator);
//			}
//		}
//
//		@Deprecated
//		public static <T extends RSDOject> void writeRSDObject(T value, BiConsumer<T,JsonGenerator> propertiesHandler, JsonGenerator generator) {
//			if( value != null ) {
//				generator.writeStartObject();
//				if( value.id() != null ) {
//					generator.write(ID, value.id());
//				}
//				if( value.etag() != null ) {
//					generator.write(ETAG, value.etag());
//				}
//				propertiesHandler.accept(value, generator);
//				generator.writeEnd();
//			} else {
//				generator.writeNull();
//			}
//		}

        public static <T> void writeArray(String keyName, Array<T> value,
                BiConsumer<T, JsonGenerator> propertiesHandler, JsonGenerator generator) {
            if (value != null) {
                generator.writeKey(keyName);
                writeArray(value, propertiesHandler, generator);
            }
        }

        public static <T> void writeArray(Array<T> value, BiConsumer<T, JsonGenerator> propertiesHandler,
                JsonGenerator generator) {
            if (value != null) {
                generator.writeStartArray();
                value.forEach(v -> propertiesHandler.accept(v, generator));
                generator.writeEnd();
            }
        }
    }
}
